"""Model(s) for streamdetails."""

from __future__ import annotations

import time
from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from typing import Any

from mashumaro import DataClassDictMixin, field_options, pass_through

from .dsp import DSPDetails
from .enums import MediaType, StreamType, VolumeNormalizationMode
from .media_items import AudioFormat

type StreamMetadataUpdateCallback = Callable[[StreamDetails, int], Awaitable[None]]


@dataclass
class StreamMetadata(DataClassDictMixin):
    """
    Metadata of a live broadcast / media stream.

    This is used, for example, to provide information about the currently playing
    track on a radio station. A provider may use this to expose additional
    metadata about the stream.
    """

    # mandatory fields
    title: str

    # optional fields
    artist: str | None = None
    album: str | None = None
    image_url: str | None = None
    duration: int | None = None
    description: str | None = None
    uri: str | None = None
    elapsed_time: int | None = None
    elapsed_time_last_updated: float | None = None  # UTC timestamp

    @property
    def corrected_elapsed_time(self) -> float | None:
        """Return the corrected/realtime elapsed time (while playing)."""
        if self.elapsed_time is None or self.elapsed_time_last_updated is None:
            return None
        return self.elapsed_time + (time.time() - self.elapsed_time_last_updated)


@dataclass
class MultiPartPath:
    """
    Model for a multipart path.

    Used when a stream is split into multiple parts, e.g. chapters.
    """

    path: str
    # use the duration field to specify the duration of this part
    # for more efficient seeking
    duration: float | None = None


@dataclass(kw_only=True)
class StreamDetails(DataClassDictMixin):
    """Model for streamdetails."""

    # NOTE: the actual provider/itemid of the streamdetails may differ
    # from the connected media_item due to track linking etc.
    # the streamdetails are only used to provide details about the content
    # that is (going to be) streamed.

    #############################################################################
    # mandatory fields                                                          #
    #############################################################################
    provider: str
    item_id: str
    audio_format: AudioFormat
    media_type: MediaType = MediaType.TRACK
    stream_type: StreamType = StreamType.CUSTOM

    #############################################################################
    # optional fields                                                           #
    #############################################################################

    # duration of the item to stream, copied from media_item if omitted
    duration: int | None = None

    # total size in bytes of the item, calculated at eof when omitted
    size: int | None = None

    # stream metadata: radio/live streams can optionally set/use this field
    # to set the metadata of the playing media during the stream
    stream_metadata: StreamMetadata | None = None

    #############################################################################
    # the fields below will only be used server-side and not sent to the client #
    #############################################################################

    # path: url or (local accessible) path to the stream (if not custom stream)
    # this field should be set by the provider when creating the streamdetails
    # unless the stream is a custom stream
    # if the stream consists of multiple parts, this may also be a list of MultiPartPath
    path: str | list[MultiPartPath] | None = field(
        default=None,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    # data: provider specific data (not exposed externally)
    # this info is for example used to pass along details to the get_audio_stream
    # this field may be set by the provider when creating the streamdetails
    data: Any = field(
        default=None,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    # extra_input_args: any additional input args to pass along to ffmpeg
    # this field may be set by the provider when creating the streamdetails
    extra_input_args: list[str] = field(
        default_factory=list,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    # decryption_key: decryption key for encrypted streams (used with StreamType.ENCRYPTED_HTTP)
    # this field may be set by the provider when creating the streamdetails
    decryption_key: str | None = field(
        default=None,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    # allow_seek: bool to indicate that the content can/may be seeked
    # If set to False, seeking will be completely disabled.
    # NOTE: this is automatically disabled for duration-less streams (e/g. radio)
    allow_seek: bool = field(
        default=False,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )

    # can_seek: bool to indicate that the custom audio stream can be seeked
    # if set to False, and allow seek is set to True, the core logic will attempt
    # to seek in the incoming (bytes)stream, which is not a guarantee it will work.
    # If allow_seek is also set to False, seeking will be completely disabled.
    can_seek: bool = field(
        default=False,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )

    # expiration: time in seconds until the streamdetails expire
    expiration: int = field(
        default=600,  # 10 minutes
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )

    # stream metadata update callback
    # optional (async) callback that will be called to update the stream metadata
    # it will be passed the streamdetails object and the elapsed time in seconds
    stream_metadata_update_callback: StreamMetadataUpdateCallback | None = field(
        default=None,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    # interval in seconds to call the stream metadata update callback
    stream_metadata_update_interval: int = field(
        default=5,  # 5 seconds
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )

    #############################################################################
    # the fields below will be set/controlled by the streamcontroller           #
    #############################################################################
    loudness: float | None = None
    loudness_album: float | None = None
    prefer_album_loudness: bool = False
    volume_normalization_mode: VolumeNormalizationMode | None = None
    volume_normalization_gain_correct: float | None = None
    target_loudness: float | None = None

    # This contains the DSPDetails of all players in the group.
    # In case of single player playback, dict will contain only one entry.
    dsp: dict[str, DSPDetails] | None = None

    # the fields below are managed by the queue/stream controller and may not be set by providers
    fade_in: bool = field(
        default=False,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    seek_position: int = field(
        default=0,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    queue_id: str | None = field(
        default=None,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    seconds_streamed: float | None = field(
        default=None,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    stream_error: bool | None = field(
        default=None,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    buffer: Any = field(  # for in-memory buffering of stream data
        default=None,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    created_at: float = field(
        default_factory=time.time,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    stream_metadata_last_updated: float | None = field(
        default=None,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )

    def __str__(self) -> str:
        """Return pretty printable string of object."""
        return self.uri

    @property
    def uri(self) -> str:
        """Return uri representation of item."""
        return f"{self.provider}://{self.media_type.value}/{self.item_id}"

    @property
    def stream_title(self) -> str | None:
        """
        Return stream title representation of item.

        Provided for backwards compatibility reasons.
        """
        if self.stream_metadata:
            if self.stream_metadata.artist:
                return f"{self.stream_metadata.artist} - {self.stream_metadata.title}"
            return self.stream_metadata.title
        return None

    @stream_title.setter
    def stream_title(self, value: str) -> None:
        """Set stream title representation of item."""
        if " - " in value:
            artist, title = value.split(" - ", 1)
            self.stream_metadata = StreamMetadata(title=title, artist=artist)
        else:
            self.stream_metadata = StreamMetadata(title=value)

    def __post_serialize__(self, d: dict[Any, Any]) -> dict[Any, Any]:
        """Execute action(s) on serialization."""
        # add alias for stream_title for backwards compatibility
        d["stream_title"] = self.stream_title
        return d



"""Custom errors and exceptions."""


class MusicAssistantError(Exception):
    """Custom Exception for all errors."""

    error_code = 0

    def __init_subclass__(cls, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        """Register a subclass."""
        super().__init_subclass__(*args, **kwargs)
        ERROR_MAP[cls.error_code] = cls


# mapping from error_code to Exception class
ERROR_MAP: dict[int, type] = {0: MusicAssistantError, 999: MusicAssistantError}


class ProviderUnavailableError(MusicAssistantError):
    """Error raised when trying to access mediaitem of unavailable provider."""

    error_code = 1


class MediaNotFoundError(MusicAssistantError):
    """Error raised when trying to access non existing media item."""

    error_code = 2


class InvalidDataError(MusicAssistantError):
    """Error raised when an object has invalid data."""

    error_code = 3


class AlreadyRegisteredError(MusicAssistantError):
    """Error raised when a duplicate music provider or player is registered."""

    error_code = 4


class SetupFailedError(MusicAssistantError):
    """Error raised when setup of a provider or player failed."""

    error_code = 5


class LoginFailed(MusicAssistantError):
    """Error raised when a login failed."""

    error_code = 6


class AudioError(MusicAssistantError):
    """Error raised when an issue arose when processing audio."""

    error_code = 7


class QueueEmpty(MusicAssistantError):
    """Error raised when trying to start queue stream while queue is empty."""

    error_code = 8


class UnsupportedFeaturedException(MusicAssistantError):
    """Error raised when a feature is not supported."""

    error_code = 9


class PlayerUnavailableError(MusicAssistantError):
    """Error raised when trying to access non-existing or unavailable player."""

    error_code = 10


class PlayerCommandFailed(MusicAssistantError):
    """Error raised when a command to a player failed execution."""

    error_code = 11


class InvalidCommand(MusicAssistantError):
    """Error raised when an unknown command is requested on the API."""

    error_code = 12


class UnplayableMediaError(MusicAssistantError):
    """Error thrown when a MediaItem cannot be played properly."""

    error_code = 13


class InvalidProviderURI(MusicAssistantError):
    """Error thrown when a provider URI does not match a known format."""

    error_code = 14


class InvalidProviderID(MusicAssistantError):
    """Error thrown when a provider media item identifier does not match a known format."""

    error_code = 15


class RetriesExhausted(MusicAssistantError):
    """Error thrown when a retries to a given provider URI have been exhausted."""

    error_code = 16


class ResourceTemporarilyUnavailable(MusicAssistantError):
    """Error thrown when a resource is temporarily unavailable."""

    def __init__(self, *args: object, backoff_time: int = 0) -> None:
        """Initialize."""
        super().__init__(*args)
        self.backoff_time = backoff_time

    error_code = 17


class ProviderPermissionDenied(MusicAssistantError):
    """Error thrown when a provider action is denied because of permissions."""

    error_code = 18


class ActionUnavailable(MusicAssistantError):
    """Error thrown when a action is denied because is is (temporary) unavailable/not possible."""

    error_code = 19


class AuthenticationRequired(MusicAssistantError):
    """Error raised when authentication is required but not provided."""

    error_code = 20


class AuthenticationFailed(MusicAssistantError):
    """Error raised when authentication credentials are invalid."""

    error_code = 21


class InsufficientPermissions(MusicAssistantError):
    """Error raised when user lacks required permissions for an action."""

    error_code = 22


class InvalidToken(MusicAssistantError):
    """Error raised when an access token is invalid or expired."""

    error_code = 23



"""Model for AudioFormat details."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from mashumaro import DataClassDictMixin

from music_assistant_models.enums import ContentType

MetadataTypes = int | bool | str | list[str]


@dataclass(kw_only=True)
class AudioFormat(DataClassDictMixin):
    """Model for AudioFormat details."""

    content_type: ContentType = ContentType.UNKNOWN
    codec_type: ContentType = ContentType.UNKNOWN
    sample_rate: int = 44100
    bit_depth: int = 16
    channels: int = 2
    output_format_str: str = ""
    bit_rate: int | None = None  # optional bitrate in kbps

    def __post_init__(self) -> None:
        """Execute actions after init."""
        if not self.output_format_str and self.content_type.is_pcm():
            self.output_format_str = (
                f"pcm;codec=pcm;rate={self.sample_rate};"
                f"bitrate={self.bit_depth};channels={self.channels}"
            )
        elif not self.output_format_str:
            self.output_format_str = self.content_type.value
        if self.bit_rate and self.bit_rate > 10000:
            # correct bit rate in bits per second to kbps
            self.bit_rate = int(self.bit_rate / 1000)

    @property
    def quality(self) -> int:
        """Calculate quality score."""
        if self.content_type.is_lossless():
            # lossless content is scored very high based on sample rate and bit depth
            return int(self.sample_rate / 1000) + self.bit_depth
        # lossy content, bit_rate is most important score
        # but prefer some codecs over others
        # calculate a rough score based on bit rate per channel
        bit_rate = self.bit_rate or 320
        bit_rate_score = (bit_rate / self.channels) / 100
        if self.content_type in (ContentType.AAC, ContentType.OGG):
            bit_rate_score += 1
        return int(bit_rate_score)

    @property
    def pcm_sample_size(self) -> int:
        """Return the PCM sample size."""
        return int(self.sample_rate * (self.bit_depth / 8) * self.channels)

    def __eq__(self, other: object) -> bool:
        """Check equality of two items."""
        if not isinstance(other, AudioFormat):
            return False
        return str(self) == str(other)

    def __str__(self) -> str:
        """Return string representation."""
        return (
            f"{self.output_format_str} {self.sample_rate}/{self.bit_depth} {self.channels} channels"
        )

    def __hash__(self) -> int:
        """Return custom hash."""
        return hash(self.__str__())

    def __post_serialize__(self, d: dict[Any, Any]) -> dict[Any, Any]:
        """Execute action(s) on serialization."""
        # bit_rate is now optional. Set default value to keep compatibility
        # TODO: remove this after release of MA 2.5
        d["bit_rate"] = d["bit_rate"] or 0
        return d



"""
Model(s) for MediaItemPlaybackProgressReport.

This data is sent with the MEDIA_ITEM_PLAYED event.
"""

from __future__ import annotations

from dataclasses import dataclass

from mashumaro import DataClassDictMixin

from .enums import MediaType


@dataclass(frozen=True)
class MediaItemPlaybackProgressReport(DataClassDictMixin):
    """Object to submit in a progress report during/after media playback."""

    # mandatory fields
    uri: str
    media_type: MediaType
    name: str
    duration: int
    seconds_played: int
    fully_played: bool
    is_playing: bool

    # optional fields
    mbid: str | None = None
    artist: str | None = None  # This is a concatenated string of all artist names
    artists: list[str] | None = None  # This is a list of the artist names
    artist_mbids: list[str] | None = None
    album: str | None = None
    album_mbid: str | None = None
    album_artist: str | None = None
    album_artist_mbids: list[str] | None = None
    image_url: str | None = None
    version: str | None = None
    userid: str | None = None  # the user(id) that initiated the playback



"""Model(s) for PlayerQueue."""

from __future__ import annotations

import time
from dataclasses import dataclass, field
from typing import Any

from mashumaro import DataClassDictMixin, field_options, pass_through

from .enums import PlaybackState, RepeatMode
from .media_items import MediaItemType
from .queue_item import QueueItem


@dataclass
class PlayLogEntry:
    """Representation of a PlayLogEntry within Music Assistant."""

    queue_item_id: str
    duration: int | None = None
    seconds_streamed: float | None = None


@dataclass
class PlayerQueue(DataClassDictMixin):
    """Representation of (the state of) a PlayerQueue within Music Assistant."""

    queue_id: str
    active: bool
    display_name: str
    available: bool
    items: int
    shuffle_enabled: bool = False
    repeat_mode: RepeatMode = RepeatMode.OFF
    dont_stop_the_music_enabled: bool = False

    # current_index: index that is active (e.g. being played) by the player
    current_index: int | None = None
    # index_in_buffer: index that has been preloaded/buffered by the player
    index_in_buffer: int | None = None

    elapsed_time: float = 0
    elapsed_time_last_updated: float = field(default_factory=time.time)
    state: PlaybackState = PlaybackState.IDLE
    current_item: QueueItem | None = None
    next_item: QueueItem | None = None
    radio_source: list[MediaItemType] = field(default_factory=list)

    flow_mode: bool = False
    resume_pos: int = 0

    #############################################################################
    # the fields below will only be used server-side and not sent to the client #
    #############################################################################

    enqueued_media_items: list[MediaItemType] = field(
        default_factory=list,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    flow_mode_stream_log: list[PlayLogEntry] = field(
        default_factory=list,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    next_item_id_enqueued: str | None = field(
        default=None,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    session_id: str | None = field(
        default=None,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    items_last_updated: float = field(
        default_factory=time.time,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )
    userid: str | None = field(
        default=None,
        compare=False,
        metadata=field_options(serialize="omit", deserialize=pass_through),
        repr=False,
    )

    @property
    def corrected_elapsed_time(self) -> float | None:
        """Return the corrected/realtime elapsed time."""
        if self.state == PlaybackState.PLAYING:
            return self.elapsed_time + (time.time() - self.elapsed_time_last_updated)
        return self.elapsed_time

    def to_cache(self) -> dict[str, Any]:
        """Return the dict that is suitable for storing into the cache db."""
        d = self.to_dict()
        d.pop("flow_mode", None)
        d.pop("current_item", None)
        d.pop("next_item", None)
        d.pop("index_in_buffer", None)
        # enqueued_media_items needs to survive a restart
        # otherwise 'dont stop the music' will not work
        d["enqueued_media_items"] = [x.to_dict() for x in self.enqueued_media_items]
        d["userid"] = self.userid
        return d



"""Model a QueueItem."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Self
from uuid import uuid4

from mashumaro import DataClassDictMixin

from .enums import MediaType
from .media_items import (
    ItemMapping,
    MediaItemImage,
    PlayableMediaItemType,
    UniqueList,
    is_track,
    media_from_dict,
)
from .streamdetails import StreamDetails


@dataclass
class QueueItem(DataClassDictMixin):
    """Representation of a queue item."""

    queue_id: str
    queue_item_id: str
    name: str
    duration: int | None
    sort_index: int = 0
    streamdetails: StreamDetails | None = None
    media_item: PlayableMediaItemType | None = None
    image: MediaItemImage | None = None
    index: int = 0
    # the available flag can be used to mark items that are not available/playable anymore
    available: bool = True

    def __post_init__(self) -> None:
        """Set default values."""
        if not self.name and self.streamdetails and self.streamdetails.stream_title:
            self.name = self.streamdetails.stream_title
        if not self.name:
            self.name = self.uri

    @property
    def uri(self) -> str:
        """Return uri for this QueueItem (for logging purposes)."""
        if self.media_item and self.media_item.uri:
            return self.media_item.uri
        return self.queue_item_id

    @property
    def media_type(self) -> MediaType:
        """Return MediaType for this QueueItem (for convenience purposes)."""
        if self.media_item:
            return self.media_item.media_type
        if self.streamdetails:
            return self.streamdetails.media_type
        return MediaType.UNKNOWN

    @classmethod
    def from_media_item(cls, queue_id: str, media_item: PlayableMediaItemType) -> QueueItem:
        """Construct QueueItem from track/radio item."""
        if is_track(media_item) and hasattr(media_item, "artists"):
            artists = "/".join(x.name for x in media_item.artists)
            name = f"{artists} - {media_item.name}"
            if media_item.version:
                name = f"{name} ({media_item.version})"
            # save a lot of data/bandwidth by simplifying nested objects
            media_item.artists = UniqueList([ItemMapping.from_item(x) for x in media_item.artists])
            if media_item.album:
                media_item.album = ItemMapping.from_item(media_item.album)
        else:
            name = media_item.name
        return cls(
            queue_id=queue_id,
            queue_item_id=uuid4().hex,
            name=name,
            duration=media_item.duration,
            media_item=media_item,
            image=get_image(media_item),
        )

    def to_cache(self) -> dict[str, Any]:
        """Return the dict that is suitable for storing into the cache db."""
        base = self.to_dict()
        base.pop("streamdetails", None)
        return base

    @classmethod
    def from_cache(cls, d: dict[Any, Any]) -> Self:
        """Restore a QueueItem from a cache dict.

        Note: We manually deserialize media_item because mashumaro doesn't correctly
        deserialize union types - it picks the first type in the union (Track)
        regardless of the actual media_type field value.
        """
        d.pop("streamdetails", None)
        # Extract and manually deserialize media_item with correct type discrimination
        media_item_dict = d.pop("media_item", None)
        result = cls.from_dict(d)
        if media_item_dict and isinstance(media_item_dict, dict):
            result.media_item = media_from_dict(media_item_dict)  # type: ignore[assignment]
        return result


def get_image(media_item: PlayableMediaItemType | None) -> MediaItemImage | None:
    """Find the Image for the MediaItem."""
    if not media_item:
        return None
    if media_item.image:
        return media_item.image
    if media_item.media_type == MediaType.TRACK and (album := getattr(media_item, "album", None)):
        return get_image(album)
    if media_item.media_type == MediaType.PODCAST_EPISODE and (
        podcast := getattr(media_item, "podcast", None)
    ):
        return get_image(podcast)
    return None



"""Common/shared (serializable) Models (dataclassses) for Music Assistant."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import Any, TypeGuard

from mashumaro import DataClassDictMixin

from music_assistant_models.enums import MediaType
from music_assistant_models.errors import InvalidDataError
from music_assistant_models.unique_list import UniqueList

from .audio_format import AudioFormat
from .media_item import (
    Album,
    Artist,
    Audiobook,
    BrowseFolder,
    Genre,
    ItemMapping,
    MediaItem,
    MediaItemType,
    PlayableMediaItemType,
    Playlist,
    Podcast,
    PodcastEpisode,
    Radio,
    RecommendationFolder,
    Track,
)
from .metadata import MediaItemChapter, MediaItemImage, MediaItemLink, MediaItemMetadata
from .provider_mapping import ProviderMapping

__all__ = [
    "Album",
    "Artist",
    "AudioFormat",
    "Audiobook",
    "BrowseFolder",
    "Genre",
    "ItemMapping",
    "MediaItem",
    "MediaItemChapter",
    "MediaItemImage",
    "MediaItemLink",
    "MediaItemMetadata",
    "MediaItemType",
    "Metadata",
    "MetadataProvider",
    "MetadataProviderStatus",
    "MetadataProviderType",
    "PlayableMediaItemType",
    "Playlist",
    "PlaylistTrack",
    "Podcast",
    "PodcastEpisode",
    "ProviderMapping",
    "Radio",
    "RecommendationFolder",
    "Track",
    "UniqueList",
]


@dataclass(kw_only=True)
class SearchResults(DataClassDictMixin):
    """Model for results from a search query."""

    artists: Sequence[Artist | ItemMapping] = field(default_factory=list)
    albums: Sequence[Album | ItemMapping] = field(default_factory=list)
    genres: Sequence[Genre | ItemMapping] = field(default_factory=list)
    tracks: Sequence[Track | ItemMapping] = field(default_factory=list)
    playlists: Sequence[Playlist | ItemMapping] = field(default_factory=list)
    radio: Sequence[Radio | ItemMapping] = field(default_factory=list)
    audiobooks: Sequence[Audiobook | ItemMapping] = field(default_factory=list)
    podcasts: Sequence[Podcast | ItemMapping] = field(default_factory=list)


def media_from_dict(media_item: dict[str, Any]) -> MediaItemType | ItemMapping:
    """Return MediaItem from dict."""
    if "provider_mappings" not in media_item:
        return ItemMapping.from_dict(media_item)
    if media_item["media_type"] == "artist":
        return Artist.from_dict(media_item)
    if media_item["media_type"] == "album":
        return Album.from_dict(media_item)
    if media_item["media_type"] == "genre":
        return Genre.from_dict(media_item)
    if media_item["media_type"] == "track":
        return Track.from_dict(media_item)
    if media_item["media_type"] == "playlist":
        return Playlist.from_dict(media_item)
    if media_item["media_type"] == "radio":
        return Radio.from_dict(media_item)
    if media_item["media_type"] == "audiobook":
        return Audiobook.from_dict(media_item)
    if media_item["media_type"] == "podcast":
        return Podcast.from_dict(media_item)
    if media_item["media_type"] == "podcast_episode":
        return PodcastEpisode.from_dict(media_item)
    raise InvalidDataError("Unknown media type")


def is_track(val: MediaItem) -> TypeGuard[Track]:
    """Return true if this MediaItem is a track."""
    return val.media_type == MediaType.TRACK



"""Authentication models for Music Assistant API."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import UTC, datetime
from enum import StrEnum
from typing import Any

from mashumaro.mixins.orjson import DataClassORJSONMixin


class UserRole(StrEnum):
    """User role enum."""

    ADMIN = "admin"
    USER = "user"


class AuthProviderType(StrEnum):
    """Authentication provider type enum."""

    BUILTIN = "builtin"
    HOME_ASSISTANT = "homeassistant"


@dataclass
class User(DataClassORJSONMixin):
    """User model."""

    user_id: str
    username: str
    role: UserRole
    enabled: bool = True
    created_at: datetime = field(default_factory=lambda: datetime.now(UTC))
    display_name: str | None = None
    avatar_url: str | None = None
    preferences: dict[str, Any] = field(default_factory=dict)
    provider_filter: list[str] = field(default_factory=list)
    player_filter: list[str] = field(default_factory=list)


@dataclass
class UserAuthProvider(DataClassORJSONMixin):
    """Link between a User and an Authentication Provider."""

    link_id: str
    user_id: str
    provider_type: AuthProviderType
    provider_user_id: str  # The user ID from the provider (e.g., HA user ID)
    created_at: datetime = field(default_factory=lambda: datetime.now(UTC))


@dataclass
class AuthToken(DataClassORJSONMixin):
    """Authentication token model."""

    token_id: str
    user_id: str
    token_hash: str
    name: str
    created_at: datetime = field(default_factory=lambda: datetime.now(UTC))
    expires_at: datetime | None = None
    last_used_at: datetime | None = None
    is_long_lived: bool = False



"""Model(s) for Player."""

from __future__ import annotations

import time
from dataclasses import dataclass, field
from typing import Any

from mashumaro import DataClassDictMixin

from .constants import PLAYER_CONTROL_NONE
from .enums import HidePlayerOption, MediaType, PlaybackState, PlayerFeature, PlayerType
from .unique_list import UniqueList

EXTRA_ATTRIBUTES_TYPES = str | int | float | bool | None


@dataclass
class DeviceInfo(DataClassDictMixin):
    """Model for a player's deviceinfo."""

    model: str = "Unknown model"
    manufacturer: str = "Unknown Manufacturer"
    software_version: str | None = None
    model_id: str | None = None
    manufacturer_id: str | None = None
    ip_address: str | None = None
    mac_address: str | None = None


@dataclass(kw_only=True)
class PlayerMedia(DataClassDictMixin):
    """Metadata of Media loading/loaded into a player."""

    uri: str  # uri or other identifier of the loaded media - mandatory!
    media_type: MediaType = MediaType.UNKNOWN
    title: str | None = None  # optional
    artist: str | None = None  # optional
    album: str | None = None  # optional
    image_url: str | None = None  # optional
    duration: int | None = None  # optional
    source_id: str | None = None  # optional (ID of the source, may be a queue id)
    queue_item_id: str | None = None  # only present for requests from queue controller
    custom_data: dict[str, Any] | None = None  # optional - must be serializable

    # optional - elapsed playback time of the currently playing media
    elapsed_time: int | None = None
    elapsed_time_last_updated: float | None = None  # UTC timestamp

    @property
    def corrected_elapsed_time(self) -> float | None:
        """Return the corrected/realtime elapsed time (while playing)."""
        if self.elapsed_time is None or self.elapsed_time_last_updated is None:
            return None
        return self.elapsed_time + (time.time() - self.elapsed_time_last_updated)


@dataclass
class PlayerSource(DataClassDictMixin):
    """Model for a player source."""

    id: str
    name: str
    # passive: this source can not be selected/activated by MA/the user
    passive: bool = False
    # can_play_pause: this source can be paused and resumed
    can_play_pause: bool = False
    # can_seek: this source can be seeked
    can_seek: bool = False
    # can_next_previous: this source can be skipped to next/previous item
    can_next_previous: bool = False

    def __hash__(self) -> int:
        """Return custom hash."""
        return hash(self.id)


@dataclass
class Player(DataClassDictMixin):
    """Representation of (the state of) a player within Music Assistant."""

    player_id: str
    provider: str  # instance_id of the player provider
    type: PlayerType
    name: str
    available: bool
    device_info: DeviceInfo
    supported_features: set[PlayerFeature] = field(default_factory=set)
    playback_state: PlaybackState = PlaybackState.IDLE
    elapsed_time: float | None = None
    elapsed_time_last_updated: float | None = None
    powered: bool | None = None
    volume_level: int | None = None
    volume_muted: bool | None = None

    # group_members: List of player group member id's or synced child`s.
    # - If this player is a dedicated group player,
    #   returns all child id's of the players in the group.
    # - If this is a syncgroup of players from the same platform (e.g. sonos),
    #   this will return the id's of players synced to this player,
    #   and this will include the player's own id (as first item in the list).
    group_members: UniqueList[str] = field(default_factory=UniqueList)

    # static_group_members: List of player group member id's that can not be ungrouped.
    static_group_members: UniqueList[str] = field(default_factory=UniqueList)

    # can_group_with: return set of player_id's this player can group/sync with
    # can also be instance id of an entire provider if all players can group with each other
    can_group_with: set[str] = field(default_factory=set)

    # synced_to: player_id of the player this player is currently synced to
    # also referred to as "sync leader"
    synced_to: str | None = None

    # active_source: return active source (id) for this player
    # this can be a player native source id as defined in 'source list'
    # or a Music Assistant queue id, if Music Assistant is the active source.
    # When set to known, the player provider has no accurate information about the source.
    # In that case, the player manager will try to find out the active source.
    active_source: str | None = None

    # source_list: return list of available (native) sources for this player
    source_list: UniqueList[PlayerSource] = field(default_factory=UniqueList)

    # active_group: return player_id of the active group for this player (if any)
    # if the player is grouped and a group is active,
    # this should be set to the group's player_id by the group player implementation.
    active_group: str | None = None

    # current_media: return current active/loaded item on the player
    # this may be a MA queue item, url, uri or some provider specific string
    # includes metadata if supported by the provider/player
    current_media: PlayerMedia | None = None

    # enabled: if the player is enabled
    # a disabled player is hidden in the UI and updates will not be processed
    # nor will it be added to the HA integration
    enabled: bool = True

    # hide_player_in_ui: if the player should be hidden in the UI
    # if set to ALWAYS, the player will be hidden in the UI
    # if set to AUTO, the player will be hidden in the UI if it's not playing
    # if set to NEVER, the player will never be hidden in the UI
    hide_player_in_ui: set[HidePlayerOption] = field(
        default_factory=lambda: {
            HidePlayerOption.WHEN_GROUP_ACTIVE,
            HidePlayerOption.WHEN_UNAVAILABLE,
            HidePlayerOption.WHEN_SYNCED,
        }
    )

    # expose_to_ha: if the player should be exposed to Home Assistant
    # if set to False, the player will not be added to the HA integration
    expose_to_ha: bool = True

    # icon: material design icon for this player
    icon: str = "mdi-speaker"

    # group_volume: if the player is a player group or syncgroup master,
    # this will return the average volume of all child players
    # if not a group player, this is just the player's volume
    group_volume: int = 100

    # extra_attributes: additional (player specific) attributes for this player
    extra_attributes: dict[str, EXTRA_ATTRIBUTES_TYPES] = field(default_factory=dict)

    # power_control: the power control attached to this player (set by config)
    power_control: str = PLAYER_CONTROL_NONE

    # volume_control: the volume control attached to this player (set by config)
    volume_control: str = PLAYER_CONTROL_NONE

    # mute_control: the volume control attached to this player (set by config)
    mute_control: str = PLAYER_CONTROL_NONE

    #############################################################################
    # helper methods and properties                                             #
    #############################################################################

    @property
    def corrected_elapsed_time(self) -> float | None:
        """Return the corrected/realtime elapsed time."""
        if self.elapsed_time is None or self.elapsed_time_last_updated is None:
            return None
        if self.playback_state == PlaybackState.PLAYING:
            return self.elapsed_time + (time.time() - self.elapsed_time_last_updated)
        return self.elapsed_time

    @property
    def current_item_id(self) -> str | None:
        """Return current_item_id from current_media (if exists)."""
        if self.current_media:
            return self.current_media.queue_item_id or self.current_media.uri
        return None

    @classmethod
    def __post_serialize__(cls, d: dict[str, Any]) -> dict[str, Any]:
        """Adjust dict object after it has been serialized."""
        # TEMP 2025-03-15: convert power to boolean for backwards compatibility
        # Remove this once the HA integration is updated to handle this
        if d["powered"] is None and d["power_control"] == PLAYER_CONTROL_NONE:
            d["powered"] = True
        # add alias for display_name for backwards compatibility
        d["display_name"] = d["name"]
        # add alias for state for backwards compatibility
        d["state"] = d["playback_state"]
        # add alias for group_childs for backwards compatibility
        d["group_childs"] = d["group_members"]
        # add alias for extra_data for backwards compatibility
        d["extra_data"] = d["extra_attributes"]
        return d

    @classmethod
    def __pre_deserialize__(cls, d: dict[Any, Any]) -> dict[Any, Any]:
        """Adjust object before it will be deserialized."""
        # add alias for playback_state for backwards compatibility
        if "playback_state" not in d and "state" in d:
            d["playback_state"] = d["state"]
        # add alias for name for backwards compatibility
        if "name" not in d and "display_name" in d:
            d["name"] = d["display_name"]
        # add alias for group_members for backwards compatibility
        if "group_members" not in d and "group_childs" in d:
            d["group_members"] = d["group_childs"]
        # add alias for extra_attributes for backwards compatibility
        if "extra_attributes" not in d and "extra_data" in d:
            d["extra_attributes"] = d["extra_data"]
        return d



"""All (global/common) constants for Music Assistant."""

SECURE_STRING_SUBSTITUTE = "this_value_is_encrypted"

PLAYER_CONTROL_NATIVE = "native"
PLAYER_CONTROL_NONE = "none"
PLAYER_CONTROL_FAKE = "fake"



"""Models for MediaItem Metadata."""

from __future__ import annotations

from dataclasses import dataclass, fields
from datetime import datetime

from mashumaro import DataClassDictMixin

from music_assistant_models.enums import ImageType, LinkType
from music_assistant_models.helpers import merge_lists
from music_assistant_models.unique_list import UniqueList


@dataclass(frozen=True, kw_only=True)
class MediaItemLink(DataClassDictMixin):
    """Model for a link."""

    type: LinkType
    url: str

    def __hash__(self) -> int:
        """Return custom hash."""
        return hash(self.type)

    def __eq__(self, other: object) -> bool:
        """Check equality of two items."""
        if not isinstance(other, MediaItemLink):
            return False
        return self.url == other.url


@dataclass(frozen=True, kw_only=True)
class MediaItemImage(DataClassDictMixin):
    """Model for a image."""

    type: ImageType
    path: str
    provider: str  # provider lookup key (only use instance id for fileproviders)
    remotely_accessible: bool = False  # url that is accessible from anywhere

    def __hash__(self) -> int:
        """Return custom hash."""
        return hash((self.type.value, self.provider, self.path))

    def __eq__(self, other: object) -> bool:
        """Check equality of two items."""
        if not isinstance(other, MediaItemImage):
            return False
        return self.__hash__() == other.__hash__()


@dataclass(frozen=True, kw_only=True)
class MediaItemChapter(DataClassDictMixin):
    """Model for a MediaItem's chapter/bookmark."""

    position: int  # sort position/number
    name: str  # friendly name
    start: float  # start position in seconds
    end: float | None = None  # start position in seconds if known

    @property
    def duration(self) -> float:
        """Return duration of chapter."""
        return self.end - self.start if self.end else 0

    def __hash__(self) -> int:
        """Return custom hash."""
        return hash(self.position)


@dataclass(kw_only=True)
class MediaItemMetadata(DataClassDictMixin):
    """Model for a MediaItem's metadata."""

    description: str | None = None
    review: str | None = None
    explicit: bool | None = None
    # NOTE: images is a list of available images, sorted by preference
    images: UniqueList[MediaItemImage] | None = None
    grouping: str | None = None
    genres: set[str] | None = None
    mood: str | None = None
    style: str | None = None
    copyright: str | None = None
    lyrics: str | None = None  # tracks only
    lrc_lyrics: str | None = None  # tracks only
    label: str | None = None
    links: set[MediaItemLink] | None = None
    performers: set[str] | None = None
    preview: str | None = None
    popularity: int | None = None
    release_date: datetime | None = None
    languages: UniqueList[str] | None = None
    # chapters is a list of available chapters, sorted by position
    # most commonly used for audiobooks and podcast episodes
    chapters: list[MediaItemChapter] | None = None
    # last_refresh: timestamp the (full) metadata was last collected
    last_refresh: int | None = None

    def update(
        self,
        new_values: MediaItemMetadata,
    ) -> MediaItemMetadata:
        """Update metadata (in-place) with new values."""
        if not new_values:
            return self
        for fld in fields(self):
            new_val = getattr(new_values, fld.name)
            if new_val is None:
                continue
            cur_val = getattr(self, fld.name)
            if isinstance(cur_val, list) and isinstance(new_val, list):
                new_val = UniqueList(merge_lists(cur_val, new_val))
                setattr(self, fld.name, new_val)
            elif isinstance(cur_val, set) and isinstance(new_val, set | list | tuple):
                cur_val.update(new_val)
            elif new_val and fld.name in (
                "popularity",
                "last_refresh",
            ):
                # some fields are always allowed to be overwritten
                # (such as popularity and last_refresh)
                setattr(self, fld.name, new_val)
            elif cur_val is None:
                setattr(self, fld.name, new_val)
        return self

    def add_image(self, image: MediaItemImage) -> None:
        """Add an image to the list."""
        if not self.images:
            self.images = UniqueList()
        self.images.append(image)



"""Generic models used for the (websockets) API communication."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from mashumaro.mixins.orjson import DataClassORJSONMixin

from .event import MassEvent
from .helpers import get_serializable_value


@dataclass
class CommandMessage(DataClassORJSONMixin):
    """Model for a Message holding a command from server to client or client to server."""

    message_id: str
    command: str
    args: dict[str, Any] | None = None


@dataclass
class ResultMessageBase(DataClassORJSONMixin):
    """Base class for a result/response of a Command Message."""

    message_id: str


@dataclass
class SuccessResultMessage(ResultMessageBase):
    """Message sent when a Command has been successfully executed."""

    result: Any = field(default=None, metadata={"serialize": lambda v: get_serializable_value(v)})
    partial: bool = False


@dataclass
class ErrorResultMessage(ResultMessageBase):
    """Message sent when a command did not execute successfully."""

    error_code: int
    details: str | None = None


# EventMessage is the same as MassEvent, this is just a alias.
EventMessage = MassEvent


@dataclass
class ServerInfoMessage(DataClassORJSONMixin):
    """Message sent by the server with it's info when a client connects."""

    server_id: str
    server_version: str
    schema_version: int
    min_supported_schema_version: int
    base_url: str
    homeassistant_addon: bool = False
    onboard_done: bool = False


MessageType = (
    CommandMessage | EventMessage | SuccessResultMessage | ErrorResultMessage | ServerInfoMessage
)


def parse_message(raw: dict[Any, Any]) -> MessageType:
    """Parse Message from raw dict object."""
    if "event" in raw:
        return EventMessage.from_dict(raw)
    if "error_code" in raw:
        return ErrorResultMessage.from_dict(raw)
    if "result" in raw:
        return SuccessResultMessage.from_dict(raw)
    if "sdk_version" in raw:
        return ServerInfoMessage.from_dict(raw)
    return CommandMessage.from_dict(raw)



"""Model for Music Assistant Event."""

from dataclasses import dataclass, field
from typing import Any

from mashumaro.mixins.orjson import DataClassORJSONMixin

from .enums import EventType
from .helpers import get_serializable_value


@dataclass(frozen=True)
class MassEvent(DataClassORJSONMixin):
    """Representation of an Event emitted in/by Music Assistant."""

    event: EventType
    object_id: str | None = None  # player_id, queue_id or uri
    data: Any = field(
        default=None,
        metadata={"serialize": lambda v: get_serializable_value(v)},
    )
